package polyglot.ext.polyj.parse;

import polyglot.ast.*;
import polyglot.visit.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.parse.*;

import polyglot.ext.polyj.ast.*;

import java.util.*;

/**
 * An <code>Access</code> represents a <code>Parse</code> of the form "P[e]"
 * where e must be an expression.  This must be an array access, although the
 * array itself is still ambiguous.
 */
public class Access extends Parse {
    Parse prefix;
    Expr index;

    public Access(BaseParser parser, Position pos, Parse prefix, Expr index) throws Exception {
        super(parser, pos);
        this.prefix = prefix;
        this.index = index;

        if (prefix instanceof Array) parser.die("Illegal array access.", pos);
    }

    public Prefix toPrefix() throws Exception {
        return toExpr();
    }

    public Receiver toReceiver() throws Exception {
        return toExpr();
    }

    public Expr toExpr() throws Exception {
        return nf.ArrayAccess(pos, prefix.toExpr(), index);
    }

    public Expr toNewArrayPrefix(Position p) throws Exception {
        return toNewArray(p);
    }

    public NewArray toNewArray(Position p, int addDims, ArrayInit init) throws Exception {
        if (prefix instanceof Name || prefix instanceof Inst) {
            // "new T.a[n]" or "new T[L,M][n]"
            List dims = new LinkedList();
            dims.add(index);
            return nf.NewArray(p, prefix.toType(), dims).additionalDims(addDims).init(init);
        }
        else if (prefix instanceof Access || prefix instanceof InstOrAccess) {
            // The prefix is "S[n]".  Recurse as if we were doing "new S[n]"
            Expr e = prefix.toNewArrayPrefix(p);

            // Take the expression we built and add a new dimension.
            if (e instanceof NewArray) {
                NewArray a = (NewArray) e;
                List dims = new LinkedList(a.dims());
                dims.add(index);
                return a.dims(dims).additionalDims(addDims).init(init);
            }
            else if (e instanceof AmbNewArray) {
                AmbNewArray a = (AmbNewArray) e;
                List dims = new LinkedList(a.dims());
                dims.add(index);
                return a.dims(dims).additionalDims(addDims).init(init);
            }
        }

        parser.die("Illegal new array expression.", p);
        return null;
    }
    
    public Parse append(List b) throws Exception {
        if (b.size() != 1) {
            parser.die("Illegal access.", pos);
        }
        Sub s = (Sub) b.get(0);
        return new Access(parser, pos, this, s.toExpr());
    }
    
    public String toString() {
        return prefix + "[" + index + "]";    
    }
    
    public Parse prepend(Parse base) throws Exception {
        return new InstOrAccess(parser, pos, base, this);
    }


    public ParamNode toParamNode() throws Exception {
        parser.die("Array accesses cannot be used as value parameters", pos);
        return null;
    }

    public Receiver toParam() throws Exception {
        // This may seem to contradict toParamNode().  The issue is that while an array
        // access itself is not final and thus cannot be a value parameter, an array
        // access *could* be a prefix of some final.  So we only allow toParam() on
        // an access that is part of another Parse being converted to a Param.
        return toExpr();
    }

}
